package jobshop;

import org.chocosolver.solver.Model;
import org.chocosolver.solver.constraints.Constraint;
import org.chocosolver.solver.variables.IntVar;
import org.chocosolver.solver.variables.SetVar;
import org.chocosolver.solver.*;
import org.chocosolver.solver.exception.SolverException;
import org.chocosolver.solver.variables.BoolVar;
import org.chocosolver.util.ESat;
import java.io.*;
import java.util.Date;
import org.chocosolver.solver.search.strategy.Search;
import org.chocosolver.solver.search.strategy.selectors.values.IntDomainMin;
import org.chocosolver.solver.search.strategy.selectors.variables.FirstFail;
import org.chocosolver.solver.search.strategy.strategy.IntStrategy;

public class Jobshop {
 public static void main(String[] args) {
  
  // les données du problème
  // -----------------------
  

  

  int n=3;   // 3 jobs
  int m=3;   // 3 machines
  int Dmax = 300; // horizon de taille 100
  
  int[][] M = new int[n][m];
  int[][] P = new int[n][m];     
  
  M[0][0]=2;  M[0][1]=1;  M[0][2]=3;
  M[1][0]=1;  M[1][1]=2;  M[1][2]=3;
  M[2][0]=2;  M[2][1]=3;  M[2][2]=1;
  
  P[0][0]=21;  P[0][1]=53;  P[0][2]=95;
  P[1][0]=21;  P[1][1]=52;  P[1][2]=16;
  P[2][0]=39;  P[2][1]=98;  P[2][2]=42;

  int[]SommeDuree = new int[n];
  SommeDuree[0]=169;
  SommeDuree[1]=89;
  SommeDuree[2]=179;
  
  int[][] Tab = new int[n][m];     
  Tab[0][0]=0;  Tab[0][1]=1;  Tab[0][2]=2;
  Tab[1][0]=3;  Tab[1][1]=4;  Tab[1][2]=5;
  Tab[2][0]=6;  Tab[2][1]=7;  Tab[2][2]=8;

  
  
  // fin lecture
  
  Model mon_modele = new Model("JobShop");       
   
  // 2. variables
  IntVar[] date_debut = new IntVar[n*m];
  for (int i=0; i<n; i++){
     for (int j=0; j<m; j++){
     int pos = Tab[i][j];
     date_debut[pos] = mon_modele.intVar("St_"+i+"_"+j, 0, Dmax);
     }
  }

  IntVar[] durees = new IntVar[n*m];
  for (int i=0; i<n; i++){
     for (int j=0; j<m; j++){
     int pos = Tab[i][j];    
     durees[pos] = mon_modele.intVar("p_"+i+"_"+j, P[i][j], P[i][j]);
     }
  }
  
  IntVar[] date_fin = new IntVar[n*m];
  for (int i=0; i<n; i++){
     for (int j=0; j<m; j++){
        int pos = Tab[i][j]; 
        date_fin[pos] = mon_modele.intVar("Ft_"+i+"_"+j, 0, Dmax);
     }
  } 

  BoolVar[][] b = new BoolVar [n*m][n*m];
  for (int i=0; i<n; i++){
     for (int j=0; j<m; j++){
        for (int u=0; u<n; u++){
            for (int v=0; v<m; v++){
                int pos1 = Tab[i][j]; 
                int pos2 = Tab[u][v]; 
                b[pos1][pos2] = mon_modele.boolVar("b_"+i+"_"+j+"_"+u+"_"+v);           
            }
        }
    }
  }  
  
  IntVar OBJ = mon_modele.intVar("Cmax", 0, 999);
  
  // contrainte 1. relation entre date de debut et de fin
  int[] coeffs = new int[3];
  coeffs[0] = 1;  // EF
  coeffs[1] = -1;  //ES
  coeffs[2] = -1; //P

  for (int i=0; i<n; i++){
     for (int j=0; j<m; j++){
     int cour = Tab[i][j];
     
       
     IntVar[] colonne = new IntVar[3]; 
     colonne[0] = date_fin[cour];
     colonne[1] = date_debut[cour];
     colonne[2] = durees[cour];
     
     mon_modele.scalar(colonne, coeffs,"=",0).post();
     }
  }  
  
  // contrainte 2 : gamme
  for (int i=0; i<n; i++){
     for (int j=1; j<m; j++){
        int prec =Tab[i][j-1];
        int cour = Tab[i][j];;
        mon_modele.arithm(date_fin[prec], "<=",date_debut[cour]).post();
     }
  }  
    
  // contrainte 3 : disjonction 
    for (int i=0; i<n-1; i++){   
	for (int j=0; j<m; j++){
            int ope_i = Tab[i][j];
            for (int u=i+1; u<n; u++){
                for (int v=0; v<m; v++){
                    int ope_u = Tab[u][v];    
			if (M[i][j]==M[u][v]){
                            mon_modele.ifThenElse( b[ope_i][ope_u]  ,
                            mon_modele.arithm(date_fin[ope_i],"<=",date_debut[ope_u]),
                            mon_modele.arithm(date_fin[ope_u],"<=",date_debut[ope_i]));
			}
                }
            }
        } 
    }
 
   
  
  // fonction objectif  
  for (int i = 0; i < n; i++){
    int valeur = i*m+m-1;
    mon_modele.arithm(date_fin[valeur], "<=",OBJ).post(); 
  }
    System.out.println("....");
    mon_modele.arithm(OBJ, "<=",Dmax).post();
    mon_modele.setObjective(Model.MINIMIZE, OBJ);
  
    FirstFail Choix_Variables = new FirstFail(mon_modele);
    IntDomainMin Choix_Domaine = new IntDomainMin();
    IntStrategy ma_strategy = Search.intVarSearch(Choix_Variables, Choix_Domaine,date_debut);

    Solver mon_solveur = mon_modele.getSolver();
    mon_solveur.setSearch(ma_strategy);
  
    Solution s = new Solution(mon_modele);  
      
    mon_solveur.showShortStatistics();
      
      Mon_message message = new Mon_message( date_debut, n, m, date_fin, b ,Tab, M); 
      mon_solveur.showDecisions(message);
      
      Date heure_debut= new Date();
      long h_debut = heure_debut.getTime();
      while (mon_solveur.solve()) {
        
        s.record();
//       int cout = s.getIntVal(OBJ);// modif Matth 23/05
        int cout =(int) mon_solveur.getBestSolutionValue();// modif Matth 23/05
        Date heure_fin = new Date();
        long h_fin= heure_fin.getTime();
        long duree = h_fin-h_debut;
        long duree_s = duree/1000;
        System.out.println("temps : " + duree_s + " s" + "    Cmax = " + cout);
        System.out.println();
      }  
  

      if (mon_solveur.isFeasible() == ESat.TRUE)
        {
//        int cout = s.getIntVal(OBJ);// modif Matth 23/05
        int cout =(int) mon_solveur.getBestSolutionValue();// modif Matth 23/05
        System.out.println("Cmax = " + cout);

        for (int i=0; i<n; i++)
        {
          for(int j=0; j<m; j++)
          {
          int position = Tab[i][j];
          int deb = s.getIntVal(date_debut[position]);
          int fin = s.getIntVal(date_fin[position]);
          System.out.print(deb + "/" + fin);
         System.out.print("    ");
         }
         System.out.println();        
       }
      }
      else
        System.out.println("pas de sol.faisable");

 }
}